"use strict";

module.exports = deployLambda;
module.exports.LambdaDeployment = LambdaDeployment;

const glob = require('glob');
const ignore = require('ignore');
const fs = require('fs');
const yazl = require('yazl');
const path = require('path');
const rawBody = require('raw-body');
const AWS = require('aws-sdk');
const crypto = require('crypto');

function deployLambda(dir, save, callback){
    console.log(typeof dir, typeof save, typeof callback);
    // optional save
    if(typeof save === 'function'){
        callback = save;
        save = false;
    }
    
    // optional dir
    else if(typeof dir === 'function'){
        callback = dir;
        save = false;
        dir = './';
    }
    
    dir = dir.split('\\').join('/');
    
    // normalize dir
    dir = (dir && typeof dir === 'string' ? dir + (dir.charAt(dir.length - 1) === '/' ? '' : '/') : './');
    
    // get all files in the dir
    glob('**', { nodir:true, cwd:dir, dot:true }, (err, matches) => {
        if(err){
            return callback(err);
        }
        
        if(matches.length === 0){
            return callback("Lambda directory not found.");
        }
        
        var i;
        
        if(~(i = matches.indexOf('.lambdaignore'))){
            matches.splice(i, 1);
            
            try {
                var ignore_files = fs.readFileSync('.lamdbaignore', 'utf8').split(/(\r\n|\r|\n)/).filter(file => file !== '');
                if(ignore_files.length){
                    matches = ignore().add(ignore_files).filter(matches);
                }
            } catch(err){
                return callback(err);
            }
        }
        
        if(!~(matches.indexOf('.lambda'))){
            return callback(new Error('Missing .lambda file'));
        }
        
        var lamda_config = JSON.parse(fs.readFileSync(path.join(dir, '.lambda'), 'utf8'));
        
        var zip = new yazl.ZipFile();
        
        matches.forEach(file => {
            zip.addFile(path.join(dir, file), file);
        });
        
        if(save === false){
            rawBody(
                zip.outputStream/*.on("error", err => callback(err))*/,
                (err, zip) => {
                    if(err){
                        return callback(err);
                    }
                    
                    (new LambdaDeployment(zip, lamda_config)).deploy(callback);
                }
            );
        } else {
            zip.outputStream.pipe(fs.createWriteStream(path.join(dir, "lambda.zip")))
                .on("error", err => {
                    callback(err);
                })
                .on("close", () => {
                    callback(null, 'saved');
                });
        }
        
            
        zip.end();
    });
}

function LambdaDeployment(zip_buffer, config){
    this.config = {
        region: "us-east-1",
        apiVersion: "2015-03-31"
    };
    this.params = {
        Description:"",
        Handler: "index.handler",
        MemorySize:128,
        Runtime: "nodejs4.3",
        Timeout: 3,
        KMSKeyArn: null
    };
    this.update = {
        Publish: true,
        ZipFile: zip_buffer
    };
    this.zip = zip_buffer;
    this.hash = crypto.createHash('sha256').update(zip_buffer).digest('base64');
    
    this.setConfig(config);
}

LambdaDeployment.prototype.deploy = function(callback){
    this.getConfig((err, config) => {
        if(err){
            return callback(err);
        }
        
        if(!config){
            return this.uploadNew(callback);
        }
        
        var update_config = false;
        
        for(var key in this.params){
            if(this.params[key] !== config[key] && key !== 'Publish'){
                update_config = true;
                break;
            }
        }
        
        if(!update_config){
            for(key in config){
                if(this.params[key] !== config[key]){
                    switch(key){
                        case 'CodeSha256':
                        case 'CodeSize':
                        case 'FunctionArn':
                        case 'Version':
                        case 'LastModified':
                            break;
                        default:
                            update_config = true;
                    }
                    if(update_config){
                        break;
                    }
                }
            }
        }
        
        if(update_config){
            this.updateConfig(callback);
        } else {
            if(config.CodeSha256 === this.hash){
                return callback(null, config);
            }
            
            this.updateFunction(callback);
        }
    });
};

LambdaDeployment.prototype.getConfig = function(callback){
    this.lambda.getFunctionConfiguration({ FunctionName: this.update.FunctionName }, (err, response) => {
        if(err && err.code === "ResourceNotFoundException"){
            return callback(null, undefined);
        }
        callback(err, response);
    });
};

LambdaDeployment.prototype.setConfig = function(config){
    for(var key in config){
        switch(key){
            case 'profile':
                this.config.credentials = new AWS.SharedIniFileCredentials({ profile:config.profile });
                break;
            case 'apiVersion':
            case 'region':
                this.config[key] = config[key];
                break;
            case 'Publish':
                this.update[key] = config[key];
                break;
            case 'FunctionName':
                this.FunctionName = config.FunctionName;
                this.update[key] = config[key];
                /* falls through */
            default:
                this.params[key] = config[key];
        }
    }
    
    this.lambda = new AWS.Lambda(this.config);
};

LambdaDeployment.prototype.updateConfig = function(callback){
    this.lambda.updateFunctionConfiguration(this.params, (err, result) => {
        if(err){
            return callback(err);
        }
        
        if(result.CodeSha256 === this.hash){
            return callback(null, result);
        }
        
        this.updateFunction(callback);
    });
};

LambdaDeployment.prototype.updateFunction = function(callback){
    this.lambda.updateFunctionCode(this.update, (err, result) => {
        if(err){
            return callback(err);
        }
        
        callback(null, result);
    });
};

LambdaDeployment.prototype.uploadNew = function(callback){
    this.params.Publish = true;
    this.params.Code = { ZipFile:this.zip };
    this.lambda.createFunction(this.params, callback);
    delete this.params.Code;
    delete this.params.Publish;
};